/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.flamenk;

import com.flamenk.article.Article;
import com.flamenk.article.ArticleExtractor;
import com.flamenk.article.manipulators.ListNodeManipulator;
import com.flamenk.article.manipulators.OnClickNodeManipulator;
import com.flamenk.article.manipulators.PolishManipulator;
import com.flamenk.article.manipulators.SocialNetworkNodeManipulator;
import com.flamenk.article.manipulators.URLManipulator;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import java.io.IOException;
import java.net.URL;

/**
 * Represents the easiest way to extract the article from a url.
 * It uses a default configuration of {@link ArticleExtractor}.
 * The {@link ArticleExtractor} is created using the following code:
 * <pre>
 * new ArticleExtractor()
 *        .addNodeManupulator(ListNodeManipulator.getInstance())
 *        .addNodeManupulator(SocialNetworkNodeManipulator.newWithDefaultSocialFilter())
 *        .addNodeManupulator(OnClickNodeManipulator.getInstance())
 *        .addNodeManupulator(URLManipulator.getInstance())
 *        .addNodeManupulator(PolishManipulator.getInstance());
 * </pre>
 * <br/><br/><b>All static methods of this implementation are Thread Safe.</b>
 * @author Torindo Nesci
 */
public final class Flamenk {
  
  private static final ArticleExtractor DEFAULT_EXTRACTOR =
          new ArticleExtractor()
          .addNodeManupulator(ListNodeManipulator.getInstance())
          .addNodeManupulator(SocialNetworkNodeManipulator.newWithDefaultSocialFilter())
          .addNodeManupulator(OnClickNodeManipulator.getInstance())
          .addNodeManupulator(URLManipulator.getInstance())
          .addNodeManupulator(PolishManipulator.getInstance());
  
  private Flamenk() {
  
  }
  
  /**
   * Extracts the article at the given url
   * using a default {@link ArticleExtractor} configuration.
   *
   * @param url The url.
   * @return The article.
   * @throws IOException If the connection cannot be established
   * with the given url.
   */
  public static Optional<Article> extractArticle(URL url)
          throws IOException {
    Preconditions.checkNotNull(url);
    
    return DEFAULT_EXTRACTOR.extractArticle(url);
  }
  
}
