/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.flamenk.article;

import com.flamenk.AlgorithmConfiguration;
import com.flamenk.TagConstants;
import com.flamenk.article.manipulators.HtmlNodeManipulator;
import com.flamenk.dom.HtmlDocument;
import com.flamenk.dom.HtmlDocumentBuilder;
import com.flamenk.dom.HtmlDocumentBuilder.HtmlDocumentWrapper;
import com.flamenk.dom.HtmlNode;
import com.flamenk.dom.HtmlNodeArticleRanker;
import static com.flamenk.dom.HtmlNodeDisplayMode.DISPLAY;
import static com.flamenk.dom.HtmlNodeDisplayMode.DISPLAY_ONLY_CHILDREN;
import static com.flamenk.dom.HtmlNodeDisplayMode.NOT_DISPLAY;
import com.flamenk.dom.HtmlNodeRange;
import com.flamenk.dom.SimpleHtmlNodeArticleRanker;
import com.flamenk.dom.SimpleTextTokenizer;
import com.flamenk.dom.TextTokenizer;
import com.flamenk.histogram.Histogram;
import com.flamenk.histogram.HistogramBuilder;
import com.flamenk.histogram.HtmlNodeTokenizer;
import com.flamenk.histogram.SimpleHtmlNodeTokenizer;
import com.flamenk.histogram.SimpleTokenRanker;
import com.flamenk.histogram.Token;
import com.flamenk.histogram.TokenRanker;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import org.apache.commons.lang.StringEscapeUtils;

/**
 * An article extractor.
 *
 * <br/><br/><b>This implementation is Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public class ArticleExtractor {

  private static final String SRC_ATTR = "src";
  private static final String ALT_ATTR = "alt";
  private static final String TITLE_ATTR = "title";
  private static final int TIMEOUT = 30000;
  private static final HtmlDocumentBuilder HTML_DOCUMENT_BUILDER =
          HtmlDocumentBuilder.getInstance();
  private volatile HtmlNodeArticleRanker mHtmlNodeRanker =
          SimpleHtmlNodeArticleRanker.getInstance();
  private volatile HtmlNodeTokenizer mNodeTokenizer =
          SimpleHtmlNodeTokenizer.getInstance();
  private volatile TextTokenizer mTextTokenizer =
          SimpleTextTokenizer.getInstance();
  private volatile TokenRanker mTokenRanker =
          SimpleTokenRanker.getInstance();
  private volatile AlgorithmConfiguration mAlgoConfig =
          AlgorithmConfiguration.getDefault();
  private volatile int mConnectionTimeout = TIMEOUT;
  private final HistogramBuilder mHistogramBuilder = HistogramBuilder.getInstance();
  private final HistogramSegmentExtractor mHistogramSegExtractor =
          HistogramSegmentExtractor.getInstance();
  private final List<HtmlNodeManipulator> mManipulators =
          new CopyOnWriteArrayList<HtmlNodeManipulator>(); //Fine written once, only used afterwards

  /**
   * Creates a new instance of {@link ArticleExtractor}.
   */
  public ArticleExtractor() {
  }

  /**
   * Adds a node manipulator.
   *
   * @param nodeManipulator The node manipulator.
   * @return This article extractor.
   */
  public ArticleExtractor addNodeManupulator(HtmlNodeManipulator nodeManipulator) {
    Preconditions.checkNotNull(nodeManipulator);

    this.mManipulators.add(nodeManipulator);
    return this;
  }

  /**
   * Sets the node article ranker.
   *
   * @param nodeRanker The node article ranker.
   * @return This article extractor.
   */
  public ArticleExtractor withNodeArticleRanker(HtmlNodeArticleRanker nodeRanker) {
    Preconditions.checkNotNull(nodeRanker);

    this.mHtmlNodeRanker = nodeRanker;
    return this;
  }

  /**
   * Sets the node tokenizer.
   *
   * @param tokenizer The node tokenizer.
   * @return This article extractor.
   */
  public ArticleExtractor withNodeTokenizer(HtmlNodeTokenizer tokenizer) {
    Preconditions.checkNotNull(tokenizer);

    this.mNodeTokenizer = tokenizer;
    return this;
  }

  /**
   * Sets the text tokenizer.
   *
   * @param tokenizer The text tokenizer.
   * @return This article extractor.
   */
  public ArticleExtractor withTextTokenizer(TextTokenizer tokenizer) {
    Preconditions.checkNotNull(tokenizer);

    this.mTextTokenizer = tokenizer;
    return this;
  }

  /**
   * Sets the token ranker.
   *
   * @param ranker The token ranker.
   * @return This article extractor.
   */
  public ArticleExtractor withTokenRanker(TokenRanker ranker) {
    Preconditions.checkNotNull(ranker);

    this.mTokenRanker = ranker;
    return this;
  }

  /**
   * Sets the http connection timeout.
   *
   * @param timeout The timeout in milliseconds.
   * @return This article extractor.
   */
  public ArticleExtractor withConnectionTimeout(int timeout) {
    Preconditions.checkArgument(timeout >= 0);

    this.mConnectionTimeout = timeout;
    return this;
  }

  /**
   * Sets the configuration parameters.
   *
   * @param configuration The configuration parameters.
   * @return This article extractor.
   */
  public ArticleExtractor withAlgorithmConfiguration(AlgorithmConfiguration configuration) {
    Preconditions.checkNotNull(configuration);

    this.mAlgoConfig = configuration;
    return this;
  }

  private void manipulateDocument(
          HtmlDocument htmlDoc,
          List<HtmlNodeRange> ranges,
          HtmlNodeRange articleRange,
          HtmlNodeRange maxRange,
          Article article) {

    Deque<HtmlNode> stack = new LinkedList<HtmlNode>();
    stack.push(htmlDoc.getRootNode());
    while (!stack.isEmpty()) {
      HtmlNode node = stack.pop();
      if (node.getDisplayMode() == NOT_DISPLAY) {
        continue;
      }
      if (!isNodeInRanges(ranges, node)) {
        node.setDisplayMode(DISPLAY_ONLY_CHILDREN);
      } else {
        node.setDisplayMode(DISPLAY);
      }

      if (articleRange.isNodeInRange(node)) {
        for (HtmlNodeManipulator manipulator : this.mManipulators) {
          manipulator.manipulate(node, article, maxRange);
        }
      }

      Iterator<HtmlNode> childrenDesceningIt =
              node.getChildren().descendingIterator();
      while (childrenDesceningIt.hasNext()) {
        stack.push(childrenDesceningIt.next());
      }
    }
  }

  private boolean isNodeInRanges(List<HtmlNodeRange> ranges, HtmlNode node) {
    for (HtmlNodeRange range : ranges) {
      HtmlNode start = range.getStartNode().get();
      HtmlNode end = range.getEndNode().get();
      if (start == end) {
        continue;
      }
      if (range.isNodeInRange(node)) {
        return true;
      }
    }

    return false;
  }

  private HtmlDocumentWrapper buildDocument(URL url) throws IOException {
    return HTML_DOCUMENT_BUILDER.build(
            this.mHtmlNodeRanker, this.mTextTokenizer, url, this.mConnectionTimeout);
  }

  /**
   * Extracts the article from the html page at the give url.
   *
   * @param url The url.
   * @return The article.
   * @throws IOException If the connection cannot be established with the given
   * url.
   */
  public Optional<Article> extractArticle(URL url) throws IOException {
    Preconditions.checkNotNull(url);

    HtmlDocumentWrapper htmlDocWrapper = buildDocument(url);
    long t1 = htmlDocWrapper.getStartProcessingTime();
    HtmlDocument htmlDoc = htmlDocWrapper.getHtmlDocument();

    Optional<HtmlNode> maxNodeOpt = htmlDoc.getNodeWithMaxRank();
    if (!maxNodeOpt.isPresent()) {
      return Optional.absent();
    }

    HtmlNode maxNode = maxNodeOpt.get();
    List<HistogramSegment> segments = buildHistogramSegments(maxNode);
    if (segments.isEmpty()) {
      return Optional.absent();
    }

    List<HtmlNodeRange> ranges = new ArrayList<HtmlNodeRange>(segments.size());
    HistogramSegment maxSegment = null;
    for (HistogramSegment segment : segments) {
      Optional<HtmlNodeRange> range = segment.getNodeRange();
      if (range.isPresent()) {
        ranges.add(range.get());
        if (maxSegment == null
                || segment.getRank() > maxSegment.getRank()) {
          maxSegment = segment;
        }
      }
    }
    if (maxSegment == null) {
      return Optional.absent();
    }

    HtmlNodeRange maxSegmentRange = maxSegment.getNodeRange().get();

    HtmlNodeRange firstRange = ranges.get(0);
    HtmlNodeRange lastRange = ranges.get(ranges.size() - 1);
    HtmlNodeRange articleWeakRange = new HtmlNodeRange()
            .withIncludedStartNode(firstRange.getStartNode().get())
            .withIncludedEndNode(lastRange.getEndNode().get());

    Article article = new Article(url);
    findTitleAndImageAndSet(articleWeakRange, maxSegmentRange, htmlDoc, article);
    manipulateDocument(htmlDoc, ranges, articleWeakRange, maxSegmentRange, article);
    String body = htmlDoc.getRootNode().toDisplayString();
    article.setBody(body);

    long processingTime = System.currentTimeMillis() - t1;
    article.setProcessingTime(processingTime);
    return Optional.of(article);
  }

  private List<HistogramSegment> buildHistogramSegments(HtmlNode node) {
    List<Token> tokens = this.mNodeTokenizer.tokenize(node);
    if (tokens.isEmpty()) {
      return Collections.EMPTY_LIST;
    }
    Histogram histogram = this.mHistogramBuilder.build(
            tokens, this.mTokenRanker, this.mAlgoConfig);
    List<HistogramSegment> segments =
            this.mHistogramSegExtractor.extract(histogram, this.mAlgoConfig);
    return segments;
  }

  //Find Title and Image
  private void findTitleAndImageAndSet(
          HtmlNodeRange articleRange,
          HtmlNodeRange maxRange,
          HtmlDocument htmlDoc,
          Article article) {

    HtmlNode titleNode = findTitleAndSet(maxRange, htmlDoc, article);
    HtmlNodeRange imgSearchRange = maxRange;
    if (titleNode != null && !maxRange.isNodeInRange(titleNode)) {
      HtmlNode endNode = articleRange.getEndNode().get();
      imgSearchRange = new HtmlNodeRange(titleNode, endNode);
    }
    findImageAndSet(imgSearchRange, htmlDoc, article);
  }

  private Multiset<String> tokenize(String str) {
    HashMultiset<String> terms = HashMultiset.create();

    String[] words = this.mTextTokenizer.tokenize(str);
    for (String word : words) {
      if (!word.trim().isEmpty()) {
        terms.add(word.toLowerCase());
      }
    }

    return terms;
  }

  private HtmlNode findTitleAndSet(
          HtmlNodeRange articleRange,
          HtmlDocument htmlDoc,
          Article article) {

    String headTitle = null;
    Optional<String> htmlTitleOpt = htmlDoc.getTitle();
    if (htmlTitleOpt.isPresent()) {
      headTitle = cleanHeadTitle(htmlTitleOpt.get());
    }

    Optional<HtmlNode> maxRankNodeOpt = htmlDoc.getNodeWithMaxRank();
    if (!maxRankNodeOpt.isPresent()) {
      if (headTitle != null) {
        article.setTitle(headTitle);
      }
      return null;
    }

    HtmlNodeRange range =
            new HtmlNodeRange()
            .withIncludedStartNode(htmlDoc.getRootNode());
    Optional<HtmlNode> startNode = articleRange.getStartNode();
    if (startNode.isPresent()) {
      range.withExcludedEndNode(startNode.get());
    }
    List<HtmlNode> titleNodes = htmlDoc.getAllNodesByName(TagConstants.H1, range);
    Optional<HtmlNode> firstH1 = htmlDoc.getFirstNodeByName(TagConstants.H1, articleRange);
    if (firstH1.isPresent()) {
      titleNodes.add(firstH1.get());
    }

    HtmlNode maxRankNode = maxRankNodeOpt.get();
    HtmlNode maxTitleNode = null;
    String maxTitle = null;
    double maxRank = 0;
    for (HtmlNode titleNode : titleNodes) {
      String title = titleNode.toText();
      double rank = rankFor(title, maxRankNode);
      if (rank > maxRank) {
        maxRank = rank;
        maxTitleNode = titleNode;
        maxTitle = title.trim();
      }
    }

    if (maxTitleNode == null && headTitle != null) {
      maxTitle = headTitle;
    } else if (headTitle != null) { //If headTitle exists and with max Rank, it's preferred
      String headTitleToRank = cleanHeadTitle(headTitle);
      double rank = rankFor(headTitleToRank, maxRankNode);
      if (rank > maxRank) {
        maxTitle = headTitleToRank;
        maxTitleNode = null;
      }
    }

    if (maxTitle != null) {
      article.setTitle(maxTitle);
    }
    if (maxTitleNode != null) {
      maxTitleNode.setDisplayMode(NOT_DISPLAY);
    }

    return maxTitleNode;
  }

  private String cleanHeadTitle(String headTitle) {
    String title = headTitle;
    if (title.contains("|")) {
      title = title.substring(0, title.lastIndexOf('|'));
      title = title.trim();
    } else if (title.contains("-")) {
      title = title.substring(0, title.lastIndexOf('-'));
      title = title.trim();
    }
    
    return title;

  }

  private void findImageAndSet(
          HtmlNodeRange articleRange,
          HtmlDocument htmlDoc,
          Article article) {
    Optional<String> titleOpt = article.getTitle();
    if (!titleOpt.isPresent()) {
      return;
    }
    String title = titleOpt.get();
    Multiset<String> titleTerms = tokenize(title);
    List<HtmlNode> images =
            htmlDoc.getAllNodesByName(TagConstants.IMG, articleRange);

    HtmlNode maxRankNode = htmlDoc.getNodeWithMaxRank().get();
    String srcImg = null;
    double maxRank = 0;
    for (HtmlNode img : images) {
      if (!img.hasAttribute(SRC_ATTR)) {
        continue;
      }
      String imgDescr = "";
      if (img.hasAttribute(ALT_ATTR)) {
        imgDescr = img.getAttributeValue(ALT_ATTR);
      }
      if (img.hasAttribute(TITLE_ATTR)) {
        imgDescr += " " + img.getAttributeValue(TITLE_ATTR);
      }
      if (!imgDescr.isEmpty()) {
        String[] imgTerms = this.mTextTokenizer.tokenize(imgDescr.toLowerCase());
        double countArticleBodyTerms = 0;
        double countArticleTitleTerms = 0;
        for (String imgTerm : imgTerms) {
          if (imgTerm.length() >= 5) {
            countArticleBodyTerms += maxRankNode.getCountForTerm(imgTerm);
            countArticleTitleTerms += titleTerms.count(imgTerm);
          }
        }

        double titleRank = 0;
        double bodyRank = 0;
        double titleDen = imgTerms.length * titleTerms.size();
        double bodyDen = imgTerms.length * maxRankNode.getNumTextTokens();
        if (titleDen > 0) {
          titleRank = countArticleTitleTerms / titleDen;
        }
        if (bodyDen > 0) {
          bodyRank = countArticleBodyTerms / bodyDen;
        }
        double rank = titleRank + bodyRank;
        if (rank > maxRank) {
          maxRank = rank;
          srcImg = img.getAttributeValue(SRC_ATTR);
        }
      }
    }

    if (srcImg != null) {
      if (!srcImg.startsWith("http://")) {
        try {
          URL imgURL = new URL(article.getUrl(), srcImg);
          srcImg = imgURL.toString();
        } catch (MalformedURLException ex) {
          //I hate empty catch blocks
          srcImg = "";
        }

      }
      article.setMainImageUrl(srcImg);
    }

  }

  private double rankFor(String title, HtmlNode maxRankNode) {
    String titleUnescaped = StringEscapeUtils.unescapeHtml(title);
    String[] terms = this.mTextTokenizer.tokenize(titleUnescaped);
    double termsCount = 0;
    for (String term : terms) {
      if (term.length() >= 5) {
        termsCount += maxRankNode.getCountForTerm(term);
      }
    }
    double rank = 0;
    if (terms.length > 0) {
      rank = termsCount / 1 + Math.log(terms.length);
    }
    return rank;
  }
}
