package com.flamenk.article;

import com.flamenk.dom.HtmlNodeRange;
import com.flamenk.histogram.Token;
import com.flamenk.util.TokenUtil;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * A histogram segment representing a sequence of contiguous histogram tokens.
 *
 * <br/><br/> <b>This implementation is not Thread Safe.</b>
 * @author Torindo Nesci
 */
public class HistogramSegment {

  private LinkedList<Token> mTokens = new LinkedList<Token>();
  private double mValue;
  private double mTextRatio;
  
  /**
   * Creates a new histogram segment.
   */
  public HistogramSegment() {
  
  }
  
  /**
   * Returns the text ratio.
   *
   * @return The text ratio.
   */
  public double getTextRatio() {
    return this.mTextRatio;
  }
  
  /**
   * Sets the segment text ratio.
   * @param textRatio The text ratio.
   */
  public void setTextRatio(double textRatio) {
    Preconditions.checkArgument(textRatio >= 0);
    
    this.mTextRatio = textRatio;
  }
  
  /**
   * Returns the rank of this histogram segment.
   *
   * @return The segment rank.
   */
  public double getRank() {
    return this.mValue;
  }
  
  /**
   * Sets the rank of this histogram segment.
   *
   * @param value The segment rank.
   */
  public void setRank(double value) {
    Preconditions.checkArgument(value >= 0);
    
    this.mValue = value;
  }
  
  /**
   * Adds the given token to this histogram segment.
   *
   * @param token The html token.
   */
  public void addToken(Token token) {
    Preconditions.checkNotNull(token);
    
    this.mTokens.add(token);
  }
  
  /**
   * Returns the size of this histogram segment in terms
   * of number of tokens.
   *
   * @return The histogram segment size.
   */
  public int size() {
    return this.mTokens.size();
  }
  
  /**
   * Returns the dom's node range of this histogram segment.
   * The closing tag tokens at the beginning or end of a segment and
   * the opening tag tokens at the end of the segment are excluded from the range.
   * If for instance the segment is made of the following tokens:
   * <br/><br/>
   * <code>
   *  &lt;/div&gt; Go to &lt;a&gt; Sevilla &lt;/a&gt;.
   *  &lt;div&gt;It's a flamenko city.&lt;/div&gt;
   *  &lt;img src="www.example.com"/&gt; &lt;p&gt;
   * </code>
   * <br/><br/>
   * The node range extremes will be the <code>Go to</code> text node
   * and the <code>img</code> node. <br/>
   * The rage may not exist is some cases as for the segments below:
   * <br/><br/>
   * <code>
   *  &lt;/div&gt; Go to &lt;p&gt;
   * </code>
   * <br/>
   * <code>
   *  &lt;/div&gt; &lt;/p&gt; &lt;/div&gt; &lt;a&gt;
   * </code>
   *
   * @return The node range.
   * @see HtmlNodeRange
   */
  public Optional<HtmlNodeRange> getNodeRange() {
    if (this.mTokens.isEmpty()) {
      return Optional.absent();
    }
    
    Token firstToken = null;
    Iterator<Token> tokensIt = this.mTokens.iterator();
    while (tokensIt.hasNext() && firstToken == null) {
      Token token = tokensIt.next();
      if (!TokenUtil.isClosingTag(token)) {
        firstToken = token;
      }
    }
    
    if (firstToken == null) {
      return Optional.absent();
    }
    
    Token lastToken = null;
    Iterator<Token> tokensInverseIt = this.mTokens.descendingIterator();
    while (tokensInverseIt.hasNext()
            && lastToken == null) {
      Token token = tokensInverseIt.next();
      if (token == firstToken) {
        break;
      }
      if (!TokenUtil.isOpeningTag(token) && !TokenUtil.isClosingTag(token)) {
        lastToken = token;
      }
    }
    
    if (lastToken == null) {
      return Optional.absent();
    }
    
    HtmlNodeRange range = new HtmlNodeRange(
            firstToken.getHtmlNode(), lastToken.getHtmlNode())
            .withStartIncluded()
            .withEndIncluded();
    return Optional.of(range);
  }
  
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder();
    for (Token token : this.mTokens) {
      builder.append(token);
      builder.append(" ");
    }
    return builder.toString();
  }
  
}
