package com.flamenk.article;

import com.flamenk.AlgorithmConfiguration;
import com.flamenk.histogram.Histogram;
import com.flamenk.histogram.HistogramBucket;
import com.flamenk.histogram.Token;
import com.flamenk.util.TokenUtil;
import com.google.common.base.Preconditions;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * A segment extractor from a histogram. It extract segments from a histogram
 * finding the start and end boundaries of the segment.
 *
 * <br/><br/><b>This implementation is Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public final class HistogramSegmentExtractor {

  private static final HistogramSegmentExtractor INSTANCE =
          new HistogramSegmentExtractor();

  private HistogramSegmentExtractor() {
  }

  /**
   * Returns the singleton instance of {@link HistogramSegmentExtractor}.
   *
   * @return The singleton instance of {@link HistogramSegmentExtractor}.
   */
  public static HistogramSegmentExtractor getInstance() {
    return INSTANCE;
  }

  /**
   * Extracts from the given histogram the segments with the text/tokens ratio
   * threshold greater than
   * {@link com.flamenk.HistogramConstants#TEXT_TOKEN_RATIO_TH}. All segments
   * with a number of buckets less than
   * {@link com.flamenk.HistogramConstants#MIN_BUCKETS_PER_SEGMENT_TH} are
   * filtered out.
   *
   * @param histogram The histogram.
   * @return The list of segments.
   */
  public List<HistogramSegment> extract(
          Histogram histogram) {
    return extract(histogram, AlgorithmConfiguration.getDefault());
  }

  /**
   * Extracts from the given histogram the segments with the text/tokens ratio
   * threshold greater than
   * {@link com.flamenk.HistogramConstants#TEXT_TOKEN_RATIO_TH}. All segments
   * with a number of buckets less than
   * {@link com.flamenk.HistogramConstants#MIN_BUCKETS_PER_SEGMENT_TH} are
   * filtered out.
   *
   * @param histogram The histogram.
   * @param conf The configuration parameters.
   * @return The list of segments.
   */
  public List<HistogramSegment> extract(
          Histogram histogram, AlgorithmConfiguration conf) {
    Preconditions.checkNotNull(histogram);

    List<HistogramSegment> segments = new LinkedList<HistogramSegment>();
    List<HistogramBucket> buckets = new LinkedList<HistogramBucket>();
    boolean segmentFound = false;
    int idxStartBlock = -1;
    int count = -1;
    for (HistogramBucket bucket : histogram) {
      count++;
      if (bucket.getSmoothValue() < conf.getMinBucketTextRatio()) {
        if (segmentFound) {
          segmentFound = false;
          if (buckets.size() >= conf.getMinSegBuckets()) {
            HistogramSegment histSegment =
                    buildSegment(idxStartBlock, buckets, histogram);
            segments.add(histSegment);
          }
        }
      } else {
        if (!segmentFound) {
          idxStartBlock = count;
          segmentFound = true;
          buckets.clear();
        }
        buckets.add(bucket);
      }
    }

    if (segmentFound && buckets.size() >= conf.getMinSegBuckets()) {
      HistogramSegment histSegment = buildSegment(idxStartBlock, buckets, histogram);
      segments.add(histSegment);
    }

    return segments;

  }

  private HistogramSegment buildSegment(
          int idxFirstBucket, List<HistogramBucket> buckets, Histogram histogram) {
    HistogramSegment segment = new HistogramSegment();

    List<Token> startTokens = extractStartTokens(idxFirstBucket, histogram);
    for (Token token : startTokens) {
      segment.addToken(token);
    }

    double segmentRank = 0;
    double segmentTextRatio = 0;
    HistogramBucket firstBucket = histogram.getBucketAt(idxFirstBucket);
    segmentRank += firstBucket.getSmoothValue();


    int count = 0;
    for (HistogramBucket bucket : buckets) {
      if (count > 0 && count < buckets.size() - 1) {
        segmentRank += bucket.getSmoothValue();
        segmentTextRatio += bucket.getTextRatio();
        for (Token token : bucket.getTokens()) {
          segment.addToken(token);
        }
      }
      count++;
    }
    if (count != 0) {
      segmentTextRatio /= count;
    } else {
      segmentTextRatio = 0;
    }

    int idxLastBucket = idxFirstBucket + buckets.size() - 1;
    HistogramBucket lastBucket = histogram.getBucketAt(idxLastBucket);
    segmentRank += lastBucket.getSmoothValue();

    List<Token> endTokens = extractEndTokens(idxLastBucket, histogram);
    for (Token token : endTokens) {
      segment.addToken(token);
    }

    segment.setRank(segmentRank);
    segment.setTextRatio(segmentTextRatio);
    return segment;

  }

  private List<Token> extractStartTokens(
          int idxFirstBucket, Histogram histogram) {
    List<Token> startCandidateTokens = new LinkedList<Token>();
    for (int i = 1; i > 0; i--) {
      if (idxFirstBucket >= i) {
        HistogramBucket prevFirstBucket =
                histogram.getBucketAt(idxFirstBucket - i);
        startCandidateTokens.addAll(prevFirstBucket.getTokens());
      }
    }

    HistogramBucket firstBucket = histogram.getBucketAt(idxFirstBucket);
    startCandidateTokens.addAll(firstBucket.getTokens());

    List<Token> startTokens =
            extractStartTokensFromCanditates(startCandidateTokens);
    return startTokens;
  }

  private static List<Token> extractStartTokensFromCanditates(
          List<Token> candidateTokens) {
    List<Token> startTokens = new LinkedList<Token>();
    Deque<Token> stack = new LinkedList<Token>();
    for (Token token : candidateTokens) {
      if (startTokens.isEmpty() && !TokenUtil.isClosingTag(token)) {
        startTokens.add(token);
      } else if (!startTokens.isEmpty()) {
        startTokens.add(token);
        if (TokenUtil.isOpeningTag(token)) {
          stack.push(token);
        } else if (TokenUtil.isClosingTag(token)) {
          if (stack.isEmpty()) {
            startTokens.clear();
          } else {
            Token prevTag = stack.pop();
            if (!prevTag.getName().equals(token.getName())) {
              startTokens.clear();
            }
          }
        }

      } //end outer else if
    }

    return startTokens;
  }

  private List<Token> extractEndTokens(
          int idxLastBucket, Histogram histogram) {

    LinkedList<Token> endCandidateTokens = new LinkedList<Token>();
    HistogramBucket lastBucket = histogram.getBucketAt(idxLastBucket);
    endCandidateTokens.addAll(lastBucket.getTokens());

    if (idxLastBucket < histogram.size() - 1) {
      HistogramBucket nextLastBucket =
              histogram.getBucketAt(idxLastBucket + 1);
      endCandidateTokens.addAll(nextLastBucket.getTokens());
    }

    return extractEndTokensFromCandidates(endCandidateTokens);
  }

  private List<Token> extractEndTokensFromCandidates(
          LinkedList<Token> candidateTokens) {
    LinkedList<Token> endTokens = new LinkedList<Token>();
    Deque<Token> stack = new LinkedList<Token>();

    Iterator<Token> tokensDescIt = candidateTokens.descendingIterator();
    while (tokensDescIt.hasNext()) {
      Token token = tokensDescIt.next();
      if (endTokens.isEmpty() && !TokenUtil.isOpeningTag(token)) {
        endTokens.addFirst(token);
      } else if (!endTokens.isEmpty()) {
        endTokens.addFirst(token);
        if (TokenUtil.isClosingTag(token)) {
          stack.push(token);
        } else if (TokenUtil.isOpeningTag(token)) {
          if (stack.isEmpty()) {
            endTokens.clear();
          } else {
            Token prevTag = stack.pop();
            if (!prevTag.getName().equals(token.getName())) {
              endTokens.clear();
            }
          }
        }

      } // end outer else if
    }

    return endTokens;
  }
  
}
