/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.flamenk.article.manipulators;

import com.flamenk.TagConstants;
import com.flamenk.article.Article;
import com.flamenk.dom.HtmlDocument;
import com.flamenk.dom.HtmlNode;
import static com.flamenk.dom.HtmlNodeDisplayMode.NOT_DISPLAY;
import com.flamenk.dom.HtmlNodeRange;
import com.flamenk.util.Consumer;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A social items manipulator. Removes the items that are used to share the info
 * on the social network such as Facebook like button, Twitter share button etc.
 * The following activities are performed:
 * <ul>
 *  <li>The social items are always removed.</li>
 * <li>If the social items
 * are in a html list, the list is removed as well along with the list header.</li>
 * <li>If the social items are after the max range, all the subsequent nodes
 * are removed as well.</li>
 * </ul>
 *
 * This manipulator can be configured adding instances of
 * {@link SocialNetworkItem}. The {@link SocialNetworkItem#getTagName()} is tested
 * against the {@link HtmlNode#getName()} and the {@link SocialNetworkItem#getAttributes()}
 * are testes against the {@link HtmlNode} attributes, the test is passed if:
 * <ul>
 *  <li>The html node name and social item tag name match.</li>
 *  <li>The html node has all the attributes defined in the social item with
 * the respective attribute values prefix matching the social item attributes value.
 * For instance, if the html node is
 * <tt>&lt;a href="http://www.facebook.com/plugins/like.php?href=flamenk.com"&gt;</tt>
 * and the social <tt>href</tt> attribute is <tt>http://www.facebook.com/plugins/like.php</tt>
 * they match because the html node <tt>href</tt> attribute starts with
 * the social <tt>href</tt> attribute.</li>
 * </ul>
 *
 * <br/><br/><b>This implementation is Thread Safe.</b>
 *
 * @see SocialNetworkItem
 * @author Torindo Nesci
 */
public class SocialNetworkNodeManipulator implements HtmlNodeManipulator {

  private static final String HREF_ATTR = "href";
  private static final String SRC_ATTR = "src";
  private static final String CLASS_ATTR = "class";
  private static final Consumer<HtmlNode> HIDE_NODE_CONSUMER =
          new Consumer<HtmlNode>() {
            public void consume(HtmlNode node) {
              node.setDisplayMode(NOT_DISPLAY);
            }
          };
  private final Map<String, List<SocialNetworkItem>> mSocialItems =
          new ConcurrentHashMap<String, List<SocialNetworkItem>>();
  private final Lock mLock = new ReentrantLock();

  /**
   * Creates a new instance of {@link SocialNetworkNodeManipulator}.
   */
  public SocialNetworkNodeManipulator() {
  }

  /**
   * Returns a new social manipulator adding to it the configuration for
   * Facebook and Twitter.
   *
   * @return A new {@link SocialNetworkNodeManipulator} with the default
   * filters.
   */
  public static SocialNetworkNodeManipulator newWithDefaultSocialFilter() {
    SocialNetworkNodeManipulator socialManipulator =
            new SocialNetworkNodeManipulator();
    socialManipulator.addFacebookItems();
    socialManipulator.addTwitterItems();
    return socialManipulator;
  }

  private void addItem(SocialNetworkItem item) {

    List<SocialNetworkItem> items = this.mSocialItems.get(item.getTagName());
    if (items == null) {
      //CopyOnWrite is fine, only few items added at the application bootstrap.
      items = new CopyOnWriteArrayList<SocialNetworkItem>();
      this.mSocialItems.put(item.getTagName(), items);
    }
    items.add(item);

  }

  private void addTwitterItems() {
    SocialNetworkItem share =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "http://twitter.com/share");
    SocialNetworkItem follow =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "http://twitter.com/")
            .withAttribute(CLASS_ATTR, "twitter-follow-button");
    SocialNetworkItem hash =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "http://twitter.com/intent/tweet?button_hashtag");
    SocialNetworkItem tweet =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "http://twitter.com/intent/tweet?");
    SocialNetworkItem shareSsl =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "https://twitter.com/share");
    SocialNetworkItem followSsl =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "https://twitter.com/")
            .withAttribute(CLASS_ATTR, "twitter-follow-button");
    SocialNetworkItem hashSsl =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "https://twitter.com/intent/tweet?button_hashtag");
    SocialNetworkItem tweetSsl =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "https://twitter.com/intent/tweet?");
    SocialNetworkItem iframe =
            new SocialNetworkItem(TagConstants.IFRAME)
            .withAttribute(SRC_ATTR, "http://platform.twitter.com/widgets");
    
    this.addItem(iframe);
    this.addItem(share);
    this.addItem(follow);
    this.addItem(hash);
    this.addItem(tweet);
    this.addItem(shareSsl);
    this.addItem(followSsl);
    this.addItem(hashSsl);
    this.addItem(tweetSsl);

  }

  private void addFacebookItems() {
    SocialNetworkItem like1 =
            new SocialNetworkItem(TagConstants.DIV)
            .withAttribute(CLASS_ATTR, TagConstants.FB_LIKE);
    SocialNetworkItem like2 =
            new SocialNetworkItem(TagConstants.FB_LIKE);
    SocialNetworkItem like3 =
            new SocialNetworkItem(TagConstants.IFRAME)
            .withAttribute(SRC_ATTR, "http://www.facebook.com/plugins/like.php?");
    SocialNetworkItem like4 =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "http://www.facebook.com/plugins/like.php?");
    SocialNetworkItem share =
            new SocialNetworkItem(TagConstants.A)
            .withAttribute(HREF_ATTR, "http://www.facebook.com/sharer.php?");

    this.addItem(like1);
    this.addItem(like2);
    this.addItem(like3);
    this.addItem(like4);
    this.addItem(share);

  }

  /**
   * Adds a social item configuration to this manipulator.
   *
   * @param item The social item configuration.
   */
  public void addSocialItem(SocialNetworkItem item) {
    Preconditions.checkNotNull(item);

    this.mLock.lock();
    try {
      this.addItem(item);
    } finally {
      this.mLock.unlock();
    }
  }

  @Override
  public void manipulate(HtmlNode node, Article article, HtmlNodeRange maxRankRange) {
    String tagName = node.getName();
    if (!this.mSocialItems.containsKey(tagName)) {
      return;
    }
    if (!isSocialNode(node)) {
      return;
    }

    Optional<HtmlNode> listNode = ListUtil.listNodeFor(node);
    if (!listNode.isPresent() && !maxRankRange.isNodeAfterRange(node)) {
      node.setDisplayMode(NOT_DISPLAY);
    } else if (!listNode.isPresent() && maxRankRange.isNodeAfterRange(node)) {
      HtmlDocument htmlDoc = node.getHtmlDocument();
      HtmlNodeRange range = new HtmlNodeRange().withIncludedStartNode(node);
      htmlDoc.consumeNodesInRange(range, HIDE_NODE_CONSUMER);
    } else if (listNode.isPresent() && !maxRankRange.isNodeAfterRange(node)) {
      listNode.get().setDisplayMode(NOT_DISPLAY);
      ListUtil.hideListHeader(listNode.get());
    } else if (listNode.isPresent() && maxRankRange.isNodeAfterRange(node)) {
      HtmlDocument htmlDoc = node.getHtmlDocument();
      HtmlNodeRange range =
              new HtmlNodeRange()
              .withIncludedStartNode(listNode.get());
      htmlDoc.consumeNodesInRange(range, HIDE_NODE_CONSUMER);
      ListUtil.hideListHeader(listNode.get());
    }

  }

  private boolean isSocialNode(HtmlNode node) {
    String tagName = node.getName();
    if (!this.mSocialItems.containsKey(tagName)) {
      return false;
    }
    List<SocialNetworkItem> socialItems = this.mSocialItems.get(tagName);
    for (SocialNetworkItem item : socialItems) {

      Iterator<Entry<String, String>> attrs =
              item.getAttributes().entrySet().iterator();
      boolean socialNode = true;
      while (attrs.hasNext() && socialNode) {
        Entry<String, String> attr = attrs.next();
        String attrName = attr.getKey();
        String attrVal = attr.getValue();
        if (!node.hasAttribute(attrName)
                || !node.getAttributeValue(attrName).startsWith(attrVal)) {
          socialNode = false;
        }
      }

      if (socialNode) {
        return true;
      }
    }

    return false;
  }
}
