/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.flamenk.article.manipulators;

import com.flamenk.TagConstants;
import com.flamenk.article.Article;
import com.flamenk.dom.HtmlNode;
import com.flamenk.dom.HtmlNodeRange;
import static com.flamenk.util.NodeUtil.hasName;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Sets the links and images urls,
 * if relative the links are translated to absolute.
 *
 * <br/><br/><b>This implementation is Thread Safe.</b>
 * @author Torindo Nesci
 */
public final class URLManipulator implements HtmlNodeManipulator {
  
  private static final URLManipulator INSTANCE = new URLManipulator();
  
  private static final String SRC_ATTR = "src";
  private static final String HREF_ATTR = "href";
  private static final String HTTP_PROTOCOL = "http://";
  private static final String HTTPS_PROTOCOL = "https://";
  private static final String ANCHOR = "#";
  
  private URLManipulator() {
  
  }
  
  /**
   * Returns the singleton instance of {@link URLManipulator}.
   *
   * @return The instance of {@link URLManipulator}.
   */
  public static URLManipulator getInstance() {
    return INSTANCE;
  }
  
  @Override
  public void manipulate(HtmlNode node, Article article, HtmlNodeRange range) {
    
    if (!hasName(node, TagConstants.A)
            && !hasName(node, TagConstants.IMG)) {
      return;
    }
    String url = null;
    if (node.hasAttribute(HREF_ATTR)) {
      url = node.getAttributeValue(HREF_ATTR);
    } else if (node.hasAttribute(SRC_ATTR)) {
      url = node.getAttributeValue(SRC_ATTR);
    }
    
    if (url == null) {
      return;
    }
    
    if (!url.startsWith(HTTP_PROTOCOL)
            && !url.startsWith(HTTPS_PROTOCOL)
            && !url.startsWith(ANCHOR)) {

      URL articleUrl = article.getUrl();
      try {
        URL objectUrl = new URL(articleUrl, url);
        url = objectUrl.toString();
      } catch (MalformedURLException ex) {
        //I hate empty catch blocks
        url = "";
      }
    }

    if (node.hasAttribute(HREF_ATTR)) {
      node.setDisplayAttribute(HREF_ATTR, url);
    } else if (node.hasAttribute(SRC_ATTR)) {
      node.setDisplayAttribute(SRC_ATTR, url);
    }
  }
}
