/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.flamenk.dom;

import com.flamenk.histogram.TagType;
import com.flamenk.util.ImmutableDeque;
import com.flamenk.util.ImmutableList;
import com.flamenk.util.NodeUtil;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import java.util.Collection;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import org.jsoup.nodes.Attribute;
import org.jsoup.nodes.Attributes;
import org.jsoup.nodes.Node;

/**
 * A node of the html document tree.
 *
 * <br/><br/> <b>This implementation is not Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public class HtmlNode implements ElementTreeDisplayable {

  private final HtmlDocument mHtmlDoc;
  private final Node mNode;
  private final Optional<HtmlNode> mParent;
  private final Deque<HtmlNode> mChildren = new LinkedList<HtmlNode>();
  //
  private final HtmlNodeMetrics mMetrics = new HtmlNodeMetrics();
  private final List<HtmlNodeMetrics> mChildBlocksMetrics = new LinkedList<HtmlNodeMetrics>();
  private final HtmlNodeMetrics mChildNoBlockMetrics = new HtmlNodeMetrics();
  //
  private HtmlNodeDisplayMode mDisplayMode = HtmlNodeDisplayMode.DISPLAY;
  private final Attributes mDiplayAttributes = new Attributes();
  private final int mIdxInDoc;
    
  HtmlNode(Node node, HtmlNode parent, int idxInDoc, HtmlDocument doc) {
    Preconditions.checkNotNull(node);
    Preconditions.checkNotNull(parent);
    Preconditions.checkNotNull(doc);

    this.mNode = node;
    this.mParent = Optional.of(parent);
    this.mHtmlDoc = doc;
    this.mIdxInDoc = idxInDoc;
    this.mDiplayAttributes.addAll(node.attributes());

    parent.mChildren.add(this);
  }

  HtmlNode(Node node, int idxInDoc, HtmlDocument doc) {
    Preconditions.checkNotNull(node);
    Preconditions.checkNotNull(doc);

    this.mNode = node;
    this.mParent = Optional.absent();
    this.mHtmlDoc = doc;
    this.mIdxInDoc = idxInDoc;
    this.mDiplayAttributes.addAll(node.attributes());
  }

  /**
   * Returns the metrics for this node relative to all html tokens.
   *
   * @return The node's metrics.
   */
  public HtmlNodeMetrics getMetrics() {
    return this.mMetrics;
  }
  
  /**
   * Returns the metrics belonging to the html tokens that are not enclosed in blocks.
   * E.g. If this node is a {@code div} node and precisely:
   * <pre>
   * {@literal
   * <div> My name is <em>John</em>.
   *   <div> Your name is <b> Mary</b>.</div>
   * </div>}
   * </pre>
   * The metrics returned are related to the tokens:
   * <pre>{@literal My name is <em>John</em>.}</pre>
   *
   * @return The metrics belonging to the tags that are enclose in blocks.
   */
  public HtmlNodeMetrics getChildNoBlockMetrics() {
    return this.mChildNoBlockMetrics;
  }
  
  /**
   * Returns the the list of metrics belonging to the children blocks.
   * In the current implementation, only the {@code div} block are taken into account,
   * however there is no guarantee that the set of blocks will not be extended
   * in the future in order to take into account not only the {@code div} blocks.
   * The returned list is immutable.
   *
   * @return The list of children's block metrics.
   */
  public List<HtmlNodeMetrics> getChildrenBlockMetrics() {
    return new ImmutableList<HtmlNodeMetrics>(this.mChildBlocksMetrics);
  }
  
  List<HtmlNodeMetrics> getMutableChildBlockMetrics() {
    return this.mChildBlocksMetrics;
  }
  
  void addChildBlockMetrics(HtmlNodeMetrics metrics) {
    Preconditions.checkNotNull(metrics);
    
    this.mChildBlocksMetrics.add(metrics);
  }
  
  void addAllChildBlockMetrics(Collection<HtmlNodeMetrics> metrics) {
    Preconditions.checkNotNull(metrics);
    
    //The block metrics are never modified.
    this.mChildBlocksMetrics.addAll(metrics);
  }
  
  int getIdxInDoc() {
    return this.mIdxInDoc;
  }
  
  Node getJsoupNode() {
    return this.mNode;
  }
  
  /**
   * Returns the html document this node belongs to.
   *
   * @return The html document this node belongs to.
   */
  public HtmlDocument getHtmlDocument() {
    return this.mHtmlDoc;
  }

  /**
   * Returns the parent of this node. If the current node is the root of tree
   * the parent does not exist.
   *
   * @return The parent of this node if exists.
   */
  public Optional<HtmlNode> getParent() {
    return this.mParent;
  }

  /**
   * Returns the number of occurrences of the given term in the current node and
   * its children.
   * This method is equivalent to {@code getMetrics().getCountForTerm(String)}.
   *
   * @param term The term.
   * @return The number occurrences.
   * @see HtmlNodeMetrics#getCountForTerm(java.lang.String)
   */
  public int getCountForTerm(String term) {
    Preconditions.checkNotNull(term);
    
    return this.mMetrics.getCountForTerm(term);
  }

  /**
   * Returns total number of tokens. E.g. If the this node is
   * <pre>
   *  {@literal <div> Hello <i> World</i>. </div>}
   * </pre> The total number of tokens is 5.
   *
   * This method is equivalent to {@code getMetrics().getNumTokens()}.
   *
   * @return The total number of tokens.
   * @see HtmlNodeMetrics#getNumTokens()
   */
  public int getNumTokens() {
    return this.mMetrics.getNumTokens();
  }

  /**
   * Returns the total number of text tokens. E.g. If the current node is
   * <pre>
   *  {@literal <div> Hello <a> World</a>. </div>}
   * </pre> The total number of tokens is 3.
   *
   * This method is equivalent to {@code getMetrics().getNumTextTokens()}.
   *
   * @return The total number of text tokens.
   * @see HtmlNodeMetrics#getNumTextTokens()
   */
  public int getNumTextTokens() {
    return this.mMetrics.getNumTextTokens();
  }

  /**
   * Returns the total number of div tokens. E.g. If the current node is
   * <pre>
   *  {@literal <div> Hello <a> World</a>. </div>}
   * </pre> The total number of div tokens is 2.
   *
   * This method is equivalent to {@code getMetrics().getNumDivTokens()}.
   *
   * @return The total number of div tokens.
   * @see HtmlNodeMetrics#getNumDivTokens()
   */
  public int getNumDivTokens() {
    return this.mMetrics.getNumDivTokens();
  }

  /**
   * Returns the total number of text tokens contained within link nodes.
   * If the node is
   * <pre>
   *  {@literal <div> Hello <a> World</a>. </div>}
   * </pre> The total number of text link tokens is 1. If the text token is
   * within an item list and a link at the same time, the token is undetermined,
   * it's counted either as link token or item list token.
   *
   * This method is equivalent to {@code getMetrics().getNumTextLinkTokens()}.
   *
   * @return The total number of text link tokens.
   * @see HtmlNodeMetrics#getNumTextLinkTokens()
   */
  public int getNumTextLinkTokens() {
    return this.mMetrics.getNumTextLinkTokens();
  }

  /**
   * Returns the total number of text tokens contained within list item nodes.
   * E.g. If the current node is
   * <pre>
   *  {@literal <div> USA cities: <li>San Diego</li> <li>Chicago</li> </div>}
   * </pre> The total number of text list item tokens is 3. If the text token is
   * within an item list and a link at the same time, the token is undetermined,
   * it's counted either as link token or item list token.
   *
   * This method is equivalent to {@code getMetrics().getNumTextLinkTokens()}.
   *
   * @return The total number of text link tokens.
   * @see HtmlNodeMetrics#getNumTextListItemTokens()
   */
  public int getNumTextListItemTokens() {
    return this.mMetrics.getNumTextListItemTokens();
  }

  /**
   * Returns the node's name. E.g. <i> #text, div, a, img, etc.</i>
   *
   * @return The node's name.
   */
  public String getName() {
    return this.mNode.nodeName();
  }

  /**
   * Returns the raw html of this node. E.g. If the current node is
   * <pre>
   *  {@literal <div> Hello <i> World</i>. </div>}
   * </pre> The raw html is
   * <pre>
   *  {@literal <div> Hello <i> World</i>. </div>}
   * </pre>
   *
   * @return The raw html.
   */
  public String getRawHtml() {
    return this.mNode.outerHtml();
  }

  /**
   * Returns the node's children.
   *
   * @return The node's children.
   */
  public Deque<HtmlNode> getChildren() {
    return new ImmutableDeque<HtmlNode>(this.mChildren);
  }

  /**
   * Tests if the node has an attribute.
   *
   * @param attrName The attribute name.
   * @return <code>true</code> if the node has the attribute, <code>false</code>
   * otherwise.
   * @throws NullPointerException If the argument is null.
   */
  public boolean hasAttribute(String attrName) {
    Preconditions.checkNotNull(attrName);

    return this.mNode.attributes().hasKey(attrName);
  }

  /**
   * Returns the attribute value given the attribute name.
   *
   * @param attrName The attribute name.
   * @return The attribute value.
   * @throws NullPointerException If the argument is null.
   */
  public String getAttributeValue(String attrName) {
    Preconditions.checkNotNull(attrName);

    return this.mNode.attributes().get(attrName);
  }

    /**
   * Returns the left sibling of this node if exists.
   *
   * @return The node's left sibling if exists.
   */
  public Optional<HtmlNode> getLeftSibling() {
    Optional<HtmlNode> parent = this.mParent;
    if (!parent.isPresent()) {
      return Optional.absent();
    }
    Iterator<HtmlNode> children = parent.get().getChildren().descendingIterator();
    while (children.hasNext()) {
      HtmlNode child = children.next();
      if (child == this && children.hasNext()) {
        HtmlNode leftSibling = children.next();
        return Optional.of(leftSibling);
      }
    }
    
    return Optional.absent();
  }

  /**
   * Returns the right sibling of this node if exists.
   *
   * @return The node's right sibling if exists.
   */
  public Optional<HtmlNode> getRightSibling() {
    Optional<HtmlNode> parent = this.getParent();
    if (!parent.isPresent()) {
      return Optional.absent();
    }
    Iterator<HtmlNode> children = parent.get().getChildren().iterator();
    while (children.hasNext()) {
      HtmlNode child = children.next();
      if (child == this && children.hasNext()) {
        HtmlNode firstRightSibling = children.next();
        return Optional.of(firstRightSibling);
      }
    }
    
    return Optional.absent();

  }
  
  /**
   * Tests if this node is descendant of the input node.
   *
   * @param node The node to be tested as ancestor of this node.
   * @return <code>true</code> if this node is descendant of the input node,
   * <code>false</code> otherwise.
   * @throws NullPointerException If the argument is null.
   */
  public boolean isDescendantOf(HtmlNode node) {
    Preconditions.checkNotNull(node);

    for (Optional<HtmlNode> parent = this.getParent();
            parent.isPresent();
            parent = parent.get().getParent()) {
      if (parent.get() == node) {
        return true;
      }
    }

    return false;
  }

  /**
   * Tests if this node comes after the input node in the html document.
   *
   * @param node The node to be test as node to come before this node.
   * @return <code>true</code> if this node comes after the input node,
   * <code>false</code> otherwise.
   * @throws NullPointerException If the argument is null.
   * @throws IllegalArgumentException if this node the input node are not part
   * of the same document tree.
   */
  public boolean isAfter(HtmlNode node) {
    Preconditions.checkNotNull(node);
    Preconditions.checkArgument(
            this.mHtmlDoc == node.getHtmlDocument());

    return this.mIdxInDoc > node.getIdxInDoc();
  }

  /**
   * Tests if this node comes before the input node in the html document.
   *
   * @param node The node to be test as node to come after this node.
   * @return <code>true</code> if this node comes before the input node,
   * <code>false</code> otherwise.
   * @throws NullPointerException If the argument is null.
   * @throws IllegalArgumentException if this node the input node are not part
   * of the same document tree.
   */
  public boolean isBefore(HtmlNode node) {
    Preconditions.checkNotNull(node);
    Preconditions.checkArgument(
            this.mHtmlDoc == node.getHtmlDocument());

    return this.mIdxInDoc < node.getIdxInDoc();

  }

  /**
   * Transforms this node into text discarding whatever is not text. The text
   * string is created on the fly visiting all the children nodes.
   *
   * @return The text contained in the node.
   */
  public String toText() {
    StringBuilder text = new StringBuilder();
    toText(this, text);
    return text.toString();

  }

  private void toText(HtmlNode node, StringBuilder text) {
    if (NodeUtil.isTextNode(node)) {
      text.append(node.getRawHtml());
    }
    for (HtmlNode childNode : node.getChildren()) {
      text.append(" ");
      toText(childNode, text);
    }
  }

  @Override
  public void setDisplayMode(HtmlNodeDisplayMode displayMode) {
    Preconditions.checkNotNull(displayMode);

    this.mDisplayMode = displayMode;
  }

  @Override
  public HtmlNodeDisplayMode getDisplayMode() {
    return this.mDisplayMode;
  }

  @Override
  public void setDisplayAttribute(String attrName, String attrValue) {
    Preconditions.checkNotNull(attrName);
    Preconditions.checkNotNull(attrValue);

    this.mDiplayAttributes.put(attrName, attrValue);
  }

  @Override
  public void removeDisplayAttribute(String attrName) {
    Preconditions.checkNotNull(attrName);

    if (this.mDiplayAttributes.hasKey(attrName)) {
      this.mDiplayAttributes.remove(attrName);
    }
  }

  @Override
  public void removeAllDisplayAttributes() {
    //Clear Doesn't work flaw int the Dataset implementation
    //this.mDiplayAttributes.dataset().clear()
    for (Attribute attr : this.mDiplayAttributes) {
      this.mDiplayAttributes.remove(attr.getKey());
    }
  }

  @Override
  public String toDisplayString() {
    StringBuilder displayString = new StringBuilder();
    toDisplayString(this, displayString);
    return displayString.toString();
  }

  private void toDisplayString(HtmlNode node, StringBuilder displayStr) {
    if (node.mDisplayMode == HtmlNodeDisplayMode.NOT_DISPLAY) {
      return;
    }

    if (NodeUtil.isContentNode(node) && node.mDisplayMode == HtmlNodeDisplayMode.DISPLAY) {
      displayStr.append(node.getRawHtml());
      return;
    }

    boolean nodeEmpty = NodeUtil.isNodeTypeEmpty(node);
    if (node.mDisplayMode == HtmlNodeDisplayMode.DISPLAY && nodeEmpty) {
      addTag(node, displayStr, TagType.EMPTY);
      return;
    }

    if (node.mDisplayMode == HtmlNodeDisplayMode.DISPLAY && !nodeEmpty) {
      addTag(node, displayStr, TagType.OPENING);
    }
    for (HtmlNode child : node.mChildren) {
      toDisplayString(child, displayStr);
    }
    if (node.mDisplayMode == HtmlNodeDisplayMode.DISPLAY && !nodeEmpty) {
      addTag(node, displayStr, TagType.CLOSING);
    }
  }

  private void addTag(HtmlNode node, StringBuilder tagBuilder, TagType tagType) {
    tagBuilder.append("<");
    if (tagType == TagType.CLOSING) {
      tagBuilder.append("/");
    }
    tagBuilder.append(node.getName());
    if (tagType == TagType.OPENING || tagType == TagType.EMPTY) {
      if (node.mDiplayAttributes.size() > 0) {
        tagBuilder.append(" ");
        tagBuilder.append(node.mDiplayAttributes.html().trim());
      }
    }
    if (tagType == TagType.EMPTY) {
      tagBuilder.append("/");
    }
    tagBuilder.append(">");
  }
}
