/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.flamenk.dom;

import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;

/**
 * The html node metrics.
 * They are the relevant info belonging to a html node needed to
 * tag a node as article, article comment, etc.
 *
 * <br/><br/> <b>This implementation is not Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public class HtmlNodeMetrics {

  private int mNumTextTokens = 0;
  private int mNumTextLinkTokens = 0;
  private int mNumTextListItemTokens = 0;
  private int mNumTokens = 0;
  private int mNumDivTokens = 0;
  
  private final Multiset<String> mTerms = HashMultiset.create();

  HtmlNodeMetrics() {
  
  }
  
  void setNumTokens(int numTokens) {
    Preconditions.checkArgument(numTokens >= 0);
    Preconditions.checkArgument(numTokens >= this.mNumTextTokens);
    Preconditions.checkArgument(numTokens >= this.mNumTextLinkTokens);
    Preconditions.checkArgument(numTokens >= this.mNumTextListItemTokens);
    Preconditions.checkArgument(numTokens >= this.mNumDivTokens);

    this.mNumTokens = numTokens;
  }
  
  void addNumTokens(int numTokens) {
    Preconditions.checkArgument(numTokens >= 0);
    
    this.mNumTokens += numTokens;
  }

  void setNumTextTokens(int numTextTokens) {
    Preconditions.checkArgument(numTextTokens >= 0);
    Preconditions.checkArgument(numTextTokens >= this.mNumTextLinkTokens);
    Preconditions.checkArgument(numTextTokens >= this.mNumTextListItemTokens);

    this.mNumTextTokens = numTextTokens;
  }
  
  void addNumTextTokens(int numTextTokens) {
    Preconditions.checkArgument(numTextTokens >= 0);

    this.mNumTextTokens += numTextTokens;
  }

  void setNumTextLinkTokens(int numTextLinkTokens) {
    Preconditions.checkArgument(numTextLinkTokens >= 0);

    this.mNumTextLinkTokens = numTextLinkTokens;
  }

  void addNumTextLinkTokens(int numTextLinkTokens) {
    Preconditions.checkArgument(numTextLinkTokens >= 0);

    this.mNumTextLinkTokens += numTextLinkTokens;
  }
  
  void setNumTextListItemTokens(int numTextListItemTokens) {
    Preconditions.checkArgument(numTextListItemTokens >= 0);

    this.mNumTextListItemTokens = numTextListItemTokens;
  }
  
  void addNumTextListItemTokens(int numTextListItemTokens) {
    Preconditions.checkArgument(numTextListItemTokens >= 0);

    this.mNumTextListItemTokens += numTextListItemTokens;
  }

  void setNumDivTokens(int numDivTokens) {
    Preconditions.checkArgument(numDivTokens >= 0);

    this.mNumDivTokens = numDivTokens;
  }
  
  void addNumDivTokens(int numDivTokens) {
    Preconditions.checkArgument(numDivTokens >= 0);

    this.mNumDivTokens += numDivTokens;
  }

  void addTerm(String term) {
    this.mTerms.add(term.toLowerCase());
  }

  /**
   * Returns the number of occurrences of the given term in this node metrics.
   *
   * @param term The term.
   * @return The number occurrences.
   */
  public int getCountForTerm(String term) {
    Preconditions.checkNotNull(term);

    return this.mTerms.count(term.toLowerCase());
  }

  /**
   * Returns total number of tokens. E.g. If the node is
   * <pre>
   *  {@literal <div> Hello <i> World</i>. </div>}
   * </pre> The total number of tokens is 5.
   *
   * @return The total number of tokens.
   */
  public int getNumTokens() {
    return this.mNumTokens;
  }

  /**
   * Returns the total number of text tokens. E.g. If the node is
   * <pre>
   *  {@literal <div> Hello <a> World</a>. </div>}
   * </pre> The total number of tokens is 3.
   *
   * @return The total number of text tokens.
   */
  public int getNumTextTokens() {
    return this.mNumTextTokens;
  }

  /**
   * Returns the total number of div tokens. E.g. If the node is
   * <pre>
   *  {@literal <div> Hello <a> World</a>. </div>}
   * </pre> The total number of div tokens is 2.
   *
   * @return The total number of div tokens.
   */
  public int getNumDivTokens() {
    return this.mNumDivTokens;
  }

  /**
   * Returns the total number of text tokens contained within link nodes.
   * If the node is
   * <pre>
   *  {@literal <div> Hello <a> World</a>. </div>}
   * </pre> The total number of text link tokens is 1. If the text token is
   * within an item list and a link at the same time, the token is undetermined,
   * it's counted either as link token or item list token.
   *
   * @return The total number of text link tokens.
   */
  public int getNumTextLinkTokens() {
    return this.mNumTextLinkTokens;
  }

  /**
   * Returns the total number of text tokens contained within list item nodes.
   * E.g. If the node is
   * <pre>
   *  {@literal <div> USA cities: <li>San Diego</li> <li>Chicago</li> </div>}
   * </pre> The total number of text list item tokens is 3. If the text token is
   * within an item list and a link at the same time, the token is undetermined,
   * it's counted either as link token or item list token.
   *
   * @return The total number of text link tokens.
   */
  public int getNumTextListItemTokens() {
    return this.mNumTextListItemTokens;
  }
  
  void add(HtmlNodeMetrics metrics) {
    Preconditions.checkNotNull(metrics);
    
    this.mNumDivTokens += metrics.getNumDivTokens();
    this.mNumTextLinkTokens += metrics.getNumTextLinkTokens();
    this.mNumTextListItemTokens += metrics.getNumTextListItemTokens();
    this.mNumTextTokens += metrics.getNumTextTokens();
    this.mNumTokens += metrics.getNumTokens();
    this.mTerms.addAll(metrics.mTerms);
  }
  
  static HtmlNodeMetrics newCopy(HtmlNodeMetrics metrics) {
    Preconditions.checkNotNull(metrics);
    
    HtmlNodeMetrics newMetrics = new HtmlNodeMetrics();
    newMetrics.mNumDivTokens = metrics.getNumDivTokens();
    newMetrics.mNumTextLinkTokens = metrics.getNumTextLinkTokens();
    newMetrics.mNumTextListItemTokens = metrics.getNumTextListItemTokens();
    newMetrics.mNumTextTokens = metrics.getNumTextTokens();
    newMetrics.mNumTokens = metrics.getNumTokens();
    newMetrics.mTerms.addAll(metrics.mTerms);
    return newMetrics;
  }
}
