/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.flamenk.dom;

import com.flamenk.TagConstants;
import static com.flamenk.util.NodeUtil.hasName;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;

/**
 * A simple html node ranker.
 * The rank is calculated for each given node's child metrics (block or no block)
 * then, all ranks are summed up in order to get the final node's rank.
 * Below the algorithm to calculate the rank for a single child metrics: <br/>
 * First the text rank is calculated.
 * <pre>
 *  - If the title is not present textRank is 1
 *  - If the node has less than 50 terms the textRank is 1
 *  - terms = number of total terms in the node text
 *  - termInNodeText = terms from the title present in the node text
 *  - textRank = 1 + 200 * termInNodeText / terms
 *  - textRank = 1 + log(tempRank)^2
 * </pre>
 * Since the text tokenizer used does not filter out the stop words,
 * all the terms less then 5 characters are skipped.
 * The text rank can be improved in the future using a scoring formula which takes
 * advantage of the terms frequency in the corpora, see
 * <a href="http://nlp.stanford.edu/IR-book/html/htmledition/tf-idf-weighting-1.html">
 * Tf-idf weighting
 * </a>. <br/>
 * The rank for the child metrics is then calculated:
 * <pre>
 *  - hybridTextTokens = textInLinks + textInListItem
 *  - textTokens = allTextTokens - hybridTextTokens * 2 / 3
 *  - tokens = allTokens + 15 * divTokens - hybridTextTokens
 *  - textRatio = textTokens / tokens
 *  - if textRatio > 0.72 then textRatio = textRatio * cos(2 * (textRatio - 0.72D))
 *  - if textRank == 1 then textRank = 0.3
 *  - child_metrics_rank = textRatio^7 * textTokens^2 * textRank
 * </pre>
 * The node rank is finally calculated:
 * <pre>
 *  - The rank is 0 if the node is neither <b>div</b> nor <b>article</b>.
 *  - The rank is 0 if the node has no tokens.
 *  - rank = 0
 *  - for each child metrics
 *  -     weight = child_metrics_text_tokens / all_tokens
 *  -     rank += child_metrics_rank * weight
 *  - end for each
 *  - rank = rank * node_text_tokens^1.6
 * </pre>
 *
 * <br/><b>This implementation is Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public final class SimpleHtmlNodeArticleRanker implements HtmlNodeArticleRanker {

  private static final SimpleHtmlNodeArticleRanker INSTANCE =
          new SimpleHtmlNodeArticleRanker();
  private static final double TH_SMOOTH = 0.72D;

  private SimpleHtmlNodeArticleRanker() {
    
  }

  /**
   * Returns the singleton instance of {@link SimpleHtmlNodeArticleRanker}.
   *
   * @return The html node article ranker.
   */
  public static SimpleHtmlNodeArticleRanker getInstance() {
    return INSTANCE;
  }

  @Override
  public double rank(HtmlNode node, TextTokenizer tokenizer) {
    Preconditions.checkNotNull(node);
    Preconditions.checkNotNull(tokenizer);

    if (!hasName(node, TagConstants.DIV)
            && !hasName(node, TagConstants.ARTICLE)) {
      return 0;
    }

    double numAllTokens = node.getNumTokens();
    if (numAllTokens == 0) {
      return 0;
    }
    
    HtmlDocument htmlDoc = node.getHtmlDocument();
    HtmlNodeMetrics childMetrics = node.getChildNoBlockMetrics();
    
    double articleRank = 0;
    double plainRank = rankMetrics(childMetrics, htmlDoc, tokenizer);
    double weight  = childMetrics.getNumTextTokens() / numAllTokens;
    
    articleRank +=  plainRank * weight;
    
    for (HtmlNodeMetrics metrics : node.getChildrenBlockMetrics()) {
      int childNumTextTokens = metrics.getNumTextTokens();
      weight =  childNumTextTokens / numAllTokens;
      plainRank = rankMetrics(metrics, htmlDoc, tokenizer);
      articleRank += plainRank * weight;
    }
    
    int numAllTextTokens = node.getNumTextTokens();
    return articleRank * Math.pow(numAllTextTokens, 1.6D);

  }
  
  private double rankMetrics(
          HtmlNodeMetrics metrics,
          HtmlDocument htmlDoc,
          TextTokenizer tokenizer) {
    
    double numHybridTextTokens =
            metrics.getNumTextLinkTokens()
            + metrics.getNumTextListItemTokens() * 2.0D / 3.0D;
    double numAllTokens =
            metrics.getNumTokens()
            + (15 * metrics.getNumDivTokens())
            - numHybridTextTokens;
    double numTextTokens =
            metrics.getNumTextTokens()
            - numHybridTextTokens;
    double textRatio = 0;
    if (numAllTokens > 0) {
      textRatio = numTextTokens / numAllTokens;
    }
    if (textRatio >= TH_SMOOTH) {
      textRatio *= Math.cos((textRatio - TH_SMOOTH) * 2.5D);
    }
    double nodeTextRank = calcNodeTextRank(metrics, htmlDoc, tokenizer);
    if (nodeTextRank == 1.0D) {
      nodeTextRank = 0.3D;
    }

    double artlceRank =
            Math.pow(textRatio, 7.0D)
            * Math.pow(numTextTokens, 2.0D)
            * nodeTextRank;
    
    return artlceRank;
  }
  
  private double calcNodeTextRank(
          HtmlNodeMetrics nodeMetrics,
          HtmlDocument htmlDoc,
          TextTokenizer tokenizer) {
    
    if (nodeMetrics.getNumTextTokens() < 50) {
      return 1;
    }
    
    Optional<String> titleOpt = htmlDoc.getTitle();
    if (!titleOpt.isPresent()) {
      return 1;
    }
    
    String title = titleOpt.get();
    if (title.contains("|")) {
      title = title.substring(0, title.indexOf('|'));
    }
    String[] titleTerms = tokenizer.tokenize(title);
    
    int count = 0;
    for (String term : titleTerms) {
      if (term.length() >= 4) {
        count += nodeMetrics.getCountForTerm(term);
      }
    }

    int numTextTokens = nodeMetrics.getNumTextTokens();
    double rank = 1 + 200.0D * count / numTextTokens;
    return 1 + Math.pow(Math.log(rank), 2);
  }

}
