package com.flamenk.histogram;

import static com.flamenk.HistogramConstants.HIST_BUCKET_NTOKENS;
import com.flamenk.TagConstants;
import com.flamenk.util.TokenUtil;
import static com.flamenk.util.TokenUtil.hasName;
import com.google.common.base.Preconditions;
import java.util.ArrayList;
import java.util.List;

/**
 * A histogram bucket.
 *
 * <br/><br/> <b>This implementation is not Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public class HistogramBucket {

  private static final TokenRanker TOKEN_RANKER =
          SimpleTokenRanker.getInstance();
  
  private final List<Token> mTokens =
          new ArrayList<Token>(HIST_BUCKET_NTOKENS);
  private double mValue = -1;
  private double mSmoothValue = -1;
  private double mTextRatio = -1;
  private TokenRanker mTokenRanker = TOKEN_RANKER;
  
  /**
   * Creates a histogram bucket.
   */
  public HistogramBucket() {
  
  }
  
  /**
   * Sets the token ranker.
   *
   * @param tokenRanker The token ranker.
   * @return This histogram bucket.
   */
  public HistogramBucket withTokenRanker(TokenRanker tokenRanker) {
    Preconditions.checkNotNull(tokenRanker);
    
    this.mTokenRanker = tokenRanker;
    return this;
  }
  
  /**
   * Sets the smoothed value of this histogram bucket.
   * A histogram bucket can have a smoothed value which
   * depends on the adjacent buckets.
   *
   * @param smoothValue The bucket's smoothed value.
   */
  public void setSmoothValue(double smoothValue) {
    Preconditions.checkArgument(smoothValue >= 0);
    
    this.mSmoothValue = smoothValue;
  }
  
  /**
   * Returns the smoothed value of this bucket.
   * A histogram bucket can have a smoothed value which
   * depends on the adjacent buckets.
   *
   * @return The buckets smoothed value.
   */
  public double getSmoothValue() {
    return this.mSmoothValue;
  }
  
  /**
   * Returns the value of this bucket.
   *
   * @return The bucket's value.
   */
  public double getValue() {
    if (this.mValue < 0) {
      this.mValue = calcValue();
    }
    return this.mValue;
  }
  
  /**
   * Returns the text ratio. It's calculated as
   * <code>number_of_text_tokens/number_of_tokens</code>. <br/>
   *
   * @return The text ratio. 0 if the number of tokens is 0.
   */
  public double getTextRatio() {
    if (this.mTextRatio < 0) {
      this.mTextRatio = calcTextRatio();
    }
    return this.mTextRatio;
  }
  
  private double calcTextRatio() {
    double value = 0;
    for (Token token : this.mTokens) {
      if (TokenUtil.isTextToken(token)
              || TokenUtil.isHeaderToken(token)
              || hasName(token, TagConstants.B)
              || hasName(token, TagConstants.I)
              || hasName(token, TagConstants.EM)
              || hasName(token, TagConstants.STRONG)
              || hasName(token, TagConstants.SPAN)) {
        value += 1;
      }
    }
    
    if (this.mTokens.isEmpty()) {
      value = 0;
    } else {
      value /= this.mTokens.size();
    }
    return value;
  }
  
  private double calcValue() {
    double value = 0;
    for (Token token : this.mTokens) {
      value += this.mTokenRanker.rank(token);
    }
    
    if (this.mTokens.isEmpty()) {
      value = 0;
    } else {
      value /= this.mTokens.size();
    }
    return value;
  }
  
  /**
   * Adds a token to this histogram bucket.
   *
   * @param node The html token.
   */
  public void addToken(Token node) {
    Preconditions.checkNotNull(node);
    
    this.mValue = -1;
    this.mTokens.add(node);
  }
  
  /**
   * Returns the list of tokens contained in this bucket.
   *
   * @return The list of tokens.
   */
  public List<Token> getTokens() {
    return this.mTokens;
  }
  
}
