/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.flamenk.histogram;

import com.flamenk.TagConstants;
import com.flamenk.dom.HtmlNode;
import com.flamenk.util.NodeUtil;
import com.google.common.base.Preconditions;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * A simple {@link HtmlNode} tokenizer.
 *
 * <br/><br/> <b>This implementation is Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public final class SimpleHtmlNodeTokenizer implements HtmlNodeTokenizer {

  private static final SimpleHtmlNodeTokenizer INSTANCE =
          new SimpleHtmlNodeTokenizer();

  private SimpleHtmlNodeTokenizer() {
  
  }
  
  /**
   * Returns the singleton instance of {@link SimpleHtmlNodeTokenizer}.
   *
   * @return The singleton instance of {@link SimpleHtmlNodeTokenizer}.
   */
  public static SimpleHtmlNodeTokenizer getInstance() {
    return INSTANCE;
  }
  
  private static class Entry {

    public HtmlNode node;
    public Token token;

    public Entry(HtmlNode node) {
      this.node = node;
    }

    public Entry(Token token) {
      this.token = token;
    }
  }

  @Override
  public List<Token> tokenize(HtmlNode htmlNode)  {
    Preconditions.checkNotNull(htmlNode);

    LinkedList<Token> tokens = new LinkedList<Token>();

    Deque<Entry> stack = new LinkedList<Entry>();
    stack.push(new Entry(htmlNode));

    while (!stack.isEmpty()) {
      Entry entry = stack.pop();
      if (entry.token != null) {
        tokens.addFirst(entry.token);
        continue;
      }

      HtmlNode node = entry.node;
      if (skipNode(node)) {
        continue;
      }
      
      if (NodeUtil.isContentNode(node)) {
        LinkedList<Token> contentTokens = buildContentTokens(node);
        Iterator<Token> contTokensIt = contentTokens.descendingIterator();
        while (contTokensIt.hasNext()) {
          Token token = contTokensIt.next();
          tokens.addFirst(token);
        }
      } else if (NodeUtil.isNodeTypeEmpty(node)) {
        Token token = new TagToken(TagType.EMPTY, node);
        tokens.addFirst(token);
      } else {
        Token tokenOpening = new TagToken(TagType.OPENING, node);
        Token tokenClose = new TagToken(TagType.CLOSING, node);
        tokens.addFirst(tokenClose);
        if (node.getChildren().isEmpty()) {
          tokens.addFirst(tokenOpening);
        } else {
          stack.push(new Entry(tokenOpening));
          for (HtmlNode child : node.getChildren()) {
            stack.push(new Entry(child));
          }
        }
      }
    }

    return tokens;
  }

  private LinkedList<Token> buildTextTokens(HtmlNode node) {
    String text = node.getRawHtml().trim();
    String[] tokens = text.split("[ ]+");
    LinkedList<Token> htmlTokens = new LinkedList<Token>();
    for (String token : tokens) {
      ContentToken htmlToken =
              new ContentToken(token, node);
      htmlTokens.add(htmlToken);
    }
    return htmlTokens;
  }

  private LinkedList<Token> buildContentTokens(HtmlNode node) {
    if (NodeUtil.hasName(node, TagConstants.TEXT)) {
      return buildTextTokens(node);
    }
    
    String content = node.getRawHtml();
    LinkedList<Token> tokens = new LinkedList<Token>();
    tokens.add(new ContentToken(content, node));

    return tokens;
  }
    
  private boolean skipNode(HtmlNode node) {
    return (NodeUtil.isContentNode(node)
            && node.getRawHtml().trim().isEmpty());
  }
}
