/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.flamenk.histogram;

import com.flamenk.dom.HtmlNode;
import com.google.common.base.Preconditions;

/**
 * A html tag token. A tag token must be one of the following
 * <table border="1">
 *  <tr>
 *     <td>Opening</td>
 *     <td>It is an opening tag: <code>{@literal <a href="www.flamenk.com">}</code></td>
 *  </tr>
 *  <tr>
 *     <td>Closing</td>
 *     <td>It's a closing tag: &lt;/a&gt;</td>
 *  </tr>
 *   <tr>
 *     <td>Empty</td>
 *     <td>It's an empty tag: <code>{@literal <img src="www.flamenk.com/logo.ico"/>}</code></td>
 *  </tr>
 *
 * </table>
 *
 * <br/><br/> <b>This implementation is Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public class TagToken extends Token {

  private final TagType mTagType;
  
  TagToken(TagType tagType, HtmlNode htmlNode) {
    super(htmlNode.getName(), htmlNode);
    
    Preconditions.checkNotNull(tagType);
    Preconditions.checkArgument(!htmlNode.getName().startsWith("#"));
    
    this.mTagType =  tagType;
  }
  
  /**
   * Tests if this tag is an opening tag.
   *
   * @return <code>true</code> if this tag is opening,
   * <code>false</code> otherwise.
   */
  public boolean isOpening() {
    return this.mTagType == TagType.OPENING;
  }
  
  /**
   * Tests if this tag is a closing tag.
   *
   * @return <code>true</code> if this tag is closing,
   * <code>false</code> otherwise.
   */
  public boolean isClosing() {
    return this.mTagType == TagType.CLOSING;
  }
  
  /**
   * Tests if this tag is an empty tag.
   *
   * @return <code>true</code> if this tag is closing,
   * <code>false</code> otherwise.
   */
  public boolean isEmpty() {
    return this.mTagType == TagType.EMPTY;
  }
  
  /**
   * Returns the string representation of this token.
   * The tag parameters (if any) are filtered out.
   *
   * @return The string representation of this token.
   */
  @Override
  public String getToken() {
    StringBuilder tagBuilder = new StringBuilder();
    tagBuilder.append("<");
    if (this.isClosing()) {
      tagBuilder.append("/");
    }
    tagBuilder.append(this.getName());
    if (this.isEmpty()) {
      tagBuilder.append("/");
    }
    tagBuilder.append(">");
    return tagBuilder.toString();
  }

}
