package com.flamenk.util;

import com.flamenk.TagConstants;
import com.flamenk.dom.HtmlNode;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;

/**
 * A collection of utility methods for {@link HtmlNode}.
 *
 * @author Torindo Nesci
 */
public final class NodeUtil {

  private NodeUtil() {
  
  }

  /**
   * Test if the input node is a content node. A content node has name
   * starting with the <i>#</i> symbol.
   *
   * @param node the html node.
   * @return <code>true</code> if the input node is a content node,
   * <code>false</code> otherwise.
   */
  public static boolean isContentNode(HtmlNode node) {
    Preconditions.checkNotNull(node);
    
    return node.getName().startsWith("#");
  }

  /**
   * Tests if the given node is a text node.
   *
   * @param node The html node.
   * @return <code>true</code> if the give node is a text node,
   * <code>false</code> otherwise.
   */
  public static boolean isTextNode(HtmlNode node) {
    Preconditions.checkNotNull(node);
    
    return TagConstants.TEXT.equals(node.getName());
  }
  
  /**
   * Tests if the given node is a text node with empty text.
   *
   * @param node The html node.
   * @return <code>true</code> if the given node is an empty text node,
   * <code>false</code> otherwise.
   */
  public static boolean isEmptyTextNode(HtmlNode node) {
    Preconditions.checkNotNull(node);
    
    return TagConstants.TEXT.equals(node.getName())
            && node.getRawHtml().trim().isEmpty();
  }
  
  /**
   * Tests if the given html name has the given name.
   *
   * @param node The html node.
   * @param name The node name.
   * @return <code>true</code> if the given html node has the given name,
   * <code>false</code> otherwise.
   */
  public static boolean hasName(HtmlNode node, String name) {
    Preconditions.checkNotNull(name);
    Preconditions.checkNotNull(node);
    
    return name.equals(node.getName());
  }
  
  /**
   * Tests is the given nodes has ancestor a node with the given
   * html name.
   *
   * @param node The html node.
   * @param ancestorName The node name.
   * @return <code>true</code> if the given node has an ancestor with
   * the given name, <code>false</code> otherwise.
   */
  public static boolean hasAncestor(HtmlNode node, String ancestorName) {
    Preconditions.checkNotNull(node);
    Preconditions.checkNotNull(ancestorName);
    
    for (Optional<HtmlNode> ancestor = node.getParent();
            ancestor.isPresent();
            ancestor = ancestor.get().getParent()) {
      if (ancestor.get().getName().equals(ancestorName)) {
        return true;
      }
    }
    
    return false;
  }
  
  /**
   * Tests if the given node is empty. An example of empty node is:
   * <pre>
   *    &lt;img src="www.flamenk.com/logo.ico"/&gt;
   * </pre>
   * Do not mistake an empty node for a node with no children as the one below:
   * <pre>
   *    &lt;a href="www.flamenk.com/logo.ico"&gt &lt;/a/&gt;
   * </pre>
   *
   * @param node The html node.
   * @return <code>true</code> if the given node is empty,
   * <code>false</code> otherwise.
   */
  public static boolean isNodeTypeEmpty(HtmlNode node) {
    Preconditions.checkNotNull(node);
    
    if (isContentNode(node)) {
      return true;
    }
    if (!node.getChildren().isEmpty()) {
      return false;
    }
    String html = node.getRawHtml();
    String closingTag = "</" + node.getName() + ">";
    if (html.lastIndexOf(closingTag) == (html.length() - closingTag.length())) {
      return false;
    }
    return true;
  }
  
}
