/*
 * Copyright 2013 Torindo Nesci.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.flamenk.util;

import com.flamenk.TagConstants;
import com.flamenk.histogram.TagToken;
import com.flamenk.histogram.Token;
import com.google.common.base.Preconditions;

/**
 * A collection of utility methods for {@link Token}.
 *
 * <br/><br/> <b>All static methods of this class are Thread Safe.</b>
 *
 * @author Torindo Nesci
 */
public final class TokenUtil {
  
  private TokenUtil() {
  
  }
  
  /**
   * Tests if the given token has the given name.
   *
   * @param token The token.
   * @param name The token name.
   * @return <code>true</code> if the given token has the given name,
   * <code>false</code> otherwise.
   */
  public static boolean hasName(Token token, String name) {
    Preconditions.checkNotNull(token);
    Preconditions.checkNotNull(name);
    
    return name.equals(token.getName());
  }
  
  /**
   * Tests if the given token is a text token.
   *
   * @param token The html token.
   * @return <code>true</code> if the token is a text token,
   * <code>false</code> otherwise.
   */
  public static boolean isTextToken(Token token) {
    Preconditions.checkNotNull(token);
    
    return TagConstants.TEXT.equals(token.getName());
  }
  
  /**
   * Tests if the given token is a header token. An header token is
   * one of the the following:
   * <lu>
   *  <li> h1 </li>
   *  <li> h2 </li>
   *  <li> h3 </li>
   *  <li> h4 </li>
   *  <li> h5 </li>
   *  <li> h6 </li>
   * </lu>
   * @param token The html token.
   * @return <code>true</code> if the token is a header token,
   * <code>false</code> otherwise.
   */
  public static boolean isHeaderToken(Token token) {
    Preconditions.checkNotNull(token);
    
    return TagConstants.H1.equals(token.getName())
            || TagConstants.H2.equals(token.getName())
            || TagConstants.H3.equals(token.getName())
            || TagConstants.H4.equals(token.getName())
            || TagConstants.H5.equals(token.getName())
            || TagConstants.H6.equals(token.getName());
  }
  
  /**
   * Tests if the given html token is a tag of html table family:
   * <lu>
   *  <li>table</li>
   *  <li>td</li>
   *  <li>tr</li>
   *  <li>th</li>
   *  <li>tbody</li>
   *  <li>thead</li>
   *  <li>tfoot</li>
   * </lu>
   *
   * @param token The html token.
   * @return <code>true</code> if the token is a table tag,
   * <code>false</code> otherwise.
   */
  public static boolean isTableToken(Token token) {
    Preconditions.checkNotNull(token);
    
    return TagConstants.TABLE.equals(token.getName())
            || TagConstants.TD.equals(token.getName())
            || TagConstants.TR.equals(token.getName())
            || TagConstants.TH.equals(token.getName())
            || TagConstants.TBODY.equals(token.getName())
            || TagConstants.THEAD.equals(token.getName())
            || TagConstants.TFOOT.equals(token.getName());
  }
  
  /**
   * Tests if the given token is an opening tag.
   *
   * @param token The html token.
   * @return <code>true</code> if the given token is an opening tag,
   * <code>false</code> otherwise.
   */
  public static boolean isOpeningTag(Token token) {
    Preconditions.checkNotNull(token);
    
    return token instanceof TagToken
            && ((TagToken) token).isOpening();
  }
  
    /**
   * Tests if the given token is a closing tag.
   *
   * @param token The html token.
   * @return <code>true</code> if the given token is a closing tag,
   * <code>false</code> otherwise.
   */
  public static boolean isClosingTag(Token token) {
    Preconditions.checkNotNull(token);
    
    return token instanceof TagToken
            && ((TagToken) token).isClosing();
  }
  
    /**
   * Tests if the given token is an empty tag.
   *
   * @param token The html token.
   * @return <code>true</code> if the given token is an empty tag,
   * <code>false</code> otherwise.
   */
  public static boolean isEmptyTag(Token token) {
    Preconditions.checkNotNull(token);
    
    return token instanceof TagToken
            && ((TagToken) token).isEmpty();
  }
}
